#!/bin/sh

gst-launch-1.0 videotestsrc num-buffers=250 ! \
  'video/x-raw,format=(string)I420,width=320,height=240,framerate=(fraction)25/1'  \
  ! avenc_mpeg4 ! queue ! mux.  audiotestsrc num-buffers=440 ! audioconvert ! \
  'audio/x-raw,rate=44100,channels=2'  ! lamemp3enc ! queue ! mux. \
  avimux name=mux ! filesink location=test_mp3.avi

gst-launch-1.0 videotestsrc num-buffers=250 ! \
  'video/x-raw,format=(string)I420,width=320,height=240,framerate=(fraction)25/1'  \
  ! avenc_mpeg4 ! queue ! mux.  audiotestsrc num-buffers=440 ! audioconvert ! \
  'audio/x-raw,rate=44100,channels=2'  ! queue ! mux. \
  avimux name=mux ! filesink location=test_raw.avi
